#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2022-2023  Nexedi SA and Contributors.
#                          Started by Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Program nxd-bom generates bill of material for a software-release or slapos node.

Usage: nxd-bom software   <path-to-installed-software>
       nxd-bom node       <slapos deploy-script>

An example of generated bill of material is provided in example/ors-bom.txt .
"""

# FIXME not scanned: git submodules, go modules, ruby gems, js packages

from __future__ import print_function

import sys, os, configparser, re, codecs
from os.path import basename, dirname, isdir, exists
from glob import glob
from collections import namedtuple
from urllib.parse import unquote
from urllib.request import urlretrieve
import email.parser
import zipfile
import shutil
from contextlib import closing
from hashlib import md5
import logging
import subprocess

from slapos.libnetworkcache import NetworkcacheClient


# PkgInfo represents information about a package
PkgInfo = namedtuple('PkgInfo', [
    'name',
    'version',
    'kind',
    'url',
    'license'])
    # TODO patches



# bom_software retrieves BOM from .installed.cfg generated by buildout along the build.
def bom_software(installed_software_path): # -> {} (name,kind) -> PkgInfo
    bom = {}
    def addbom(urlpath, kind, version=None):
        #print('addbom', urlpath)
        license = None      # XXX temp
        name, ver = namever(urlpath)
        if version is not None:
            assert ver is None
            ver = version
        if ver is None:
            # some packages come without version in the url
            if name in ('libstemmer_c',):
                ver = ''
            else:
                assert ver is not None, urlpath
        ver = removeprefix(ver, name+'-')   # wendelin.core-2.0.alpha3-0-g6315384  -> 2.0.alpha3-0-g6315384
        if '//' in urlpath:
            url = urlpath
            license = license_of(url)
        else:
            if kind == 'egg':
                # XXX not strictly correct -> better retrieve the actual URL, but buildout does not save it in installed.cfg
                # remove +slapospatcheXXX suffix from egg url
                v = ver
                m = re.match(r'(?P<ver>.*)\+slapospatched.*', ver, re.I)
                if m is not None:
                    v = m.group('ver')
                # if ver comes with nxd -> use nexedi.org instead of pypi
                if 'nxd' in v.lower()  or  'slapos' in v.lower():
                    nxdbase = 'http://www.nexedi.org/static/packages/source'
                    if name == 'zc.buildout':
                        nxdbase += '/slapos.buildout'
                    url = '%s/%s-%s.tar.gz' % (nxdbase, name, v)
                else:
                    url = 'https://pypi.org/project/%s/%s/' % (name, v)

                if isdir(urlpath):
                    with open('%s/EGG-INFO/PKG-INFO' % urlpath, 'rb') as f:
                        pkginfo = f.read()
                else:
                    # unzip (e.g. setuptools)
                    with zipfile.ZipFile(urlpath, 'r') as z:
                        with z.open('EGG-INFO/PKG-INFO') as f:
                            pkginfo = f.read()
                p = email.parser.BytesParser()
                meta = p.parsebytes(pkginfo, headersonly=True)
                license = meta['license']
            else:
                raise NotImplementedError('TODO url for kind %r  (urlpath: %r)' % (kind, urlpath))

        if license and license.lower() == 'unknown':
            license = None
        if license is None:     # XXX temp
            license = 'GPL-compatible'
        license = spdx_license_canon(license)
        info = PkgInfo(name, ver, kind, url, license)
        bkey = (name, kind)
        if bkey in bom:
            assert bom[bkey] == info,  (bom[bkey], info)
        else:
            bom[bkey] = info


    _ = {
        '__buildout_space__':   ' ',
        '__buildout_space_n__': '\n',
        '__buildout_space_r__': '\r',
        '__buildout_space_f__': '\f',
        '__buildout_space_v__': '\v',
    }
    idb = configparser.ConfigParser(defaults=_)
    installed_cfg = '%s/.installed.cfg' % installed_software_path
    _ = idb.read(installed_cfg)
    if not _:
        raise RuntimeError("Cannot load %r" % installed_cfg)

    for s in idb.sections():
        if s == 'buildout':
            continue    # [buildout] is used internally

        part = idb[s]
        recipe = part['recipe']

        if recipe == 'slapos.recipe.cmmi':
            addbom(geturl(part), '')    # XXX detect kind?

        elif recipe == 'slapos.recipe.build':
            # slapos.recipe.build is often used in creative ways to actually
            # run python code during the build. Let's detect this via lack of
            # `url`, but be careful and explicitly list name of such sections
            # not to miss a normal part.
            url = geturl(part, None)
            if url is None:
                # [gcc] is generic section to either use system gcc, or install one via slapos
                # the one installed via slapos will come in its own section
                # [python] is similar - it picks up  python2 or python3
                # ZODB, ZEO, ... are similar as well
                if s in ('gcc', 'python', 'ZODB', 'ZEO', 'tempstorage', 'generic_testrunner_init'):
                    continue

                raise NotImplementedError('%s uses %s without url' % (s, recipe))

            else:
                addbom(url, '')     # XXX detect kind?

        elif recipe in ('slapos.recipe.build:download', 'slapos.recipe.build:download-unpacked'):
            url = geturl(part)

            # strip off hash from the url
            vhash = None
            h = url.rfind('#')
            if h != -1:
                url, vhash = url[:h], url[h+1:]
                vhash = vhash[:8]   # abbreviate

            # slapos.recipe.build:download is often used to download .conf files, but sometimes it is used to download e.g. binaries
            # skip the part, if we can detect that downloaded item is a configuration file
            if isconf(url):
                continue

            # if url refer to local filesystem and comes as /base/... - it
            # comes from the SR itself and should be ignored
            base = part.get('_profile_base_location_')      # not always present, e.g. no on
            if base is not None  and  url.startswith(base): # ocropy-eng-traineddata
                continue

            # let's see if maybe its an archive from known place
            _ = namever(url, failonerr=False)
            if _ is not None:
                addbom(url, '', vhash)  # XXX detect kind
                continue

            raise NotImplementedError('%s uses %s with url that does not look like a .conf, archive or SR file: %s' % (s, recipe, url))


        elif recipe.startswith('slapos.recipe.template') or \
             recipe == 'collective.recipe.template':
            url = geturl(part, None)
            if url is not None:
                if isconf(url):
                    continue
                # NOTE binary is not expected with slapos.recipe.template
                raise ValueError('%s uses %s with url that does not look like a .conf file: %s' % (s, recipe, url))
            else:
                # it is an inline= script
                if recipe.startswith('slapos'):
                    assert 'inline' in part, part
                else: # collective...
                    assert part['input'].startswith('inline:')


        elif recipe in ('zc.recipe.egg:custom',  'zc.recipe.egg:develop'):
            eggpath = part['__buildout_installed__']
            assert len(eggpath.split()) == 1, eggpath  # no spaces inside - just one item
            isegg  = eggpath.endswith('.egg')
            islink = eggpath.endswith('.egg-link'),
            assert isegg or islink, eggpath
            if isegg:                   # ignore .egg-link - we declare it through the place
                addbom(eggpath, 'egg')  # from where destination is downloaded from

        elif recipe in ('zc.recipe.egg', 'zc.recipe.egg:eggs', 'zc.recipe.egg:scripts'):
            # zc.recipe.egg:* installs
            #   1) eggs, that are explicitly specified,
            #   2) indirect dependencies of eggs from "1", and
            #   3) scripts that use all those eggs from "1" and "2".
            # go through installed scripts and retrieve the eggs we can
            # retrieve to cover both "1" and "2". This does not cover the case
            # when no scripts are installed at all, but is better than nothing.
            # An alternative is to fix zc.recipe.egg itself to emit information
            # about all eggs it installs:
            xeggs = {} # xeggname -> /path/to/xeggname-ver-....egg on the filesystem
            installedv = part['__buildout_installed__'].split()
            for f in installedv:
                for xeggpath in eggscript_imports(f):
                    if not xeggpath.endswith('.egg'):
                        continue    # e.g. neoppod-repository, *.egg-link
                    xegg, _ = namever(xeggpath)
                    if xegg in xeggs:
                        assert xeggs[xegg] == xeggpath, (xeggs[xegg], xeggpath)
                    else:
                        xeggs[xegg] = xeggpath
                        addbom(xeggpath, 'egg')

            # now go through explicitly listed eggs and also add them
            # XXX sadly zc.recipe.egg neither saves in .installed.cfg information about where the eggs are installed,
            #     nor there are [versions] save.
            #     -> look the egg on the filesystem
            # TODO it is better to fix zc.recipe.egg to save the full information, so that we can build BOM just from .installed.cfg
            eggdir = part['_e']
            eggdev = part['_d']
            eggs = part.get('eggs', part.name).split('\n')
            for eggname in eggs:
                if eggname == '':   # happens when %(__buildout_space_n__)s is appended  to eggname
                    continue        # because software.cfg contained vspace
                m = _egg_re.match(eggname)
                assert m is not None, eggname
                eggname = m.group('name')   # neoppod[admin, ctl, master] -> neoppod
                eggname = eggcanon(eggname) # cython-zstd -> cython_zstd

                if eggname in xeggs:
                    continue    # we already reported this egg via scripts above

                if glob('%s/%s.egg-link' % (eggdev, eggname)):
                    continue # e.g. wendelin.core.egg-link points to wendelin.core cloned via git and already reported as git

                installed_eggs = glob('%s/*.egg' % eggdir) + \
                                 glob('%s/*.egg' % eggdev)
                eggv = [_ for _ in installed_eggs if basename(_).lower().startswith('%s-' % eggname)]
                if len(eggv) == 0:
                    raise ValueError('egg %s not found' % eggname)
                if len(eggv) > 1:
                    eggv.sort()
                    raise ValueError('egg %s is present multiple times: %s' % (eggname, eggv))
                addbom(eggv[0], 'egg')

        elif recipe == 'slapos.recipe.build:gitclone':
            repo = part['repository']
            ver  = part.get('revision')
            if ver is None:
                ver = part.get('branch')
            if ver is None:
                ver = 'HEAD'
            addbom(repo, 'git', ver)

        elif recipe in ('plone.recipe.command',):
            # not downloading anything
            pass

        else:
            raise NotImplementedError('TODO: add support for recipe %s' % recipe)


    return bom


# geturl returns value of part:url
_missing = object()
def geturl(part, default=_missing):
    url = part.get('url', raw=True)     # raw because e.g. .../ZABBIX%20Latest%20Stable/...
    if url is None:
        if default is not _missing:
            return default
        raise KeyError('section %s has no url' % part)
    return url

_egg_re = re.compile(r'^(?P<name>[\w\-\.]+)(\[.*\])?$')

# eggcanon returns canonical name for an egg.
# e.g. Cython-Zstd -> cython_zstd
def eggcanon(eggname):
    return eggname.lower().replace('-', '_')


# eggscript_imports returns list of eggs put into python path in python script
# generated by zc.recipe.egg at path.
_re_importv = re.compile(r"\nsys\.path\[0:0\] = \[\n(?P<importv>( *'[^'\n]+',\n)+) *]", re.MULTILINE)
def eggscript_imports(path):
    with open(path, 'r') as f:
        data = f.read()

    m = _re_importv.search(data)
    if m is None:
        raise RuntimeError("%s does not look like zc.recipe.egg script" % path)

    importv = []
    for x in m.group('importv').splitlines():
        x = x.strip()
        x = removeprefix(x, "'")
        x = removesuffix(x, "',")
        x = codecs.escape_decode(x)[0].decode('utf-8')
        importv.append(x)

    return importv


def bom_node(XXX):
    1/0
    # TODO bom_node should:
    #
    # 1. retrieve list of software that slapos-node brings with itself
    #    since slapos-node is built via buildout this could be extracted via
    #    node_software from /opt/slapos/.installed.cfg once we fix slapos-node
    #    package to preserve that .installed.cfg in the package itself.
    #
    # 2. retrieve list of base-OS packages that are installed by https://deploy.rapid.space/slapos
    #    This can be computed as the difference in between two sets of base-OS packages:
    #    b) set of base-OS packages after deploy.rapid.space/slapos was run, and
    #    a) set of base-OS packages after minimal base-OS was installed.
    #
    #    this computation could be automated via
    #    - preparing e.g. Debian 11 minimal chroot via debootstrap,
    #    - taking snapshot for list of installed packages via `dpkg -l`,
    #    - then deploying slapos-node in the chroot,
    #    - taking snapshot for the list of installed package again,
    #    - and finally computing the difference.
    #
    # Please see example/ors-bom.txt for "BOM(slapos-node)", "List of software
    # that slapos-node brings with itself" and "List of base-OS packages
    # installed by https://deploy.rapid.space/slapos" sections for what end result could look like.



# namever extracts item name and version from an url/path.
# for example:
#
#     http://www.python.org/ftp/python/2.7.18/Python-2.7.18.tar.xz  ->  ('Python', '2.7.18')
_gitweb_re = re.compile(r'/gitweb/\?p=(?P<name>\w+)\.git;a=snapshot;h=(?P<rev>\w+)')
_github_re = re.compile(r'/github.com/[\w\-]+/(?P<name>[\w\-]+)/archive/(refs/tags/)?(?P<rev>.+)$')
_github_rre= re.compile(r'/raw.githubusercontent.com/[\w\-]+/(?P<name>[\w\-]+)/(?P<rev>[\w\.\-]+)/')
_gitlab_re = re.compile(r'/lab.nexedi.com/[\w\-]+/(?P<name>[\w\-]+)/repository/archive.tar.(gz|bz2)\?ref=(?P<rev>.+)$')
_SF_re     = re.compile(r'/sourceforge.net/.+/(?P<name>[\w\_]+)-(?P<rev>[\w\.]+)/download$')
_git_re    = re.compile(r'/(?P<name>[\w\.\-]+)\.git$')
_go_re     = re.compile(r'/golang.org/dl/(?P<name>go)(?P<rev>[\w\.]+).src$')
_go_dev_re = re.compile(r'/go.dev/dl/(?P<name>go)(?P<rev>[\w\.]+).src$')
_osdn_f    = re.compile(r'/osdn.net/frs/redir.php\?f=(?P<f>.+)$')
def namever(url, failonerr=True): # -> (name, ver)  | None if !failonerr
    # http://www.ijg.org/files/jpegsrc.v9d.tar.gz  ->  http://www.ijg.org/files/jpegsrc.v9d
    def del_tgztail(s):
        for tail in ('.tgz', '.tar.gz', '.tbz', '.tar.bz2', '.tar.xz', '.tar.lz', '.zip', '.egg'):
            s = removesuffix(s, tail)
        for tail in ('-linux-x86_64', '_Linux_x86-64_rpm'): # FIXME x86/arm, non-linux ?
            s = removesuffix(s, tail)
        for tail in ['-py2.7'] + ['-py3.%d' % _ for _ in range(20)]:
            s = removesuffix(s, tail)
        for tail in ('-source', '-src', '.orig'):
            s = removesuffix(s, tail)
        return s
    url = del_tgztail(url)

    _ = _namever(url, failonerr)
    if _ is None:
        assert not failonerr
        return None

    name, ver = _
    if ver is not None:
        # swig-3.0.12.tar.gz  ->  swig-3.0.12
        # originally from https://sourceforge.net/projects/swig/files/swig/swig-3.0.12/swig-3.0.12.tar.gz/download
        ver = del_tgztail(ver)
        ver = removeprefix(ver, 'v')
    # XXX hack: ERP5 SR uses both readline8 and readline-5
    # XXX       Cloudooo SR ----//---- libpng16 and libpng12
    # XXX       Cloudooo SR ----//---- python2 and python3
    if name == 'readline' and ver.startswith('5'):    name += '5'
    if name == 'libpng'   and ver.startswith('1.2'):  name += '12'
    if name == 'Python'   and ver.startswith('3'):    name += '3'
    return name, ver

def _namever(url, failonerr):
    for r in (_gitweb_re, _github_re, _github_rre, _gitlab_re, _SF_re, _go_re, _go_dev_re):
        m = r.search(url)
        if m is not None:
            return m.group('name'), m.group('rev')

    m = _git_re.search(url)
    if m is not None:
        return m.group('name'), None

    m = _osdn_f.search(url)
    if m is not None:
        url = unquote(m.group('f'))

    filename = basename(url)

    name   = filename
    ver    = ''
    dashed = False
    while 1:
        # re.rsplit([-_], name, 1)
        m = re.search('[-_][^-_]*$', name)
        if m is None:
            break
        dashed = True
        v = name[m.start():]
        v_ndigit = 0
        for _ in v:
            if _.isnumeric():
                v_ndigit += 1
        isver = 'patch' in v            or \
                'slapospatched' in v    or \
                v_ndigit > (len(v)-1) / 3
        if 'rfc' in v:  # ex strict_rfc3339-0.7
            isver = False
        if not isver:   # doesn't look like a version
            break
        name = name[:m.start()]
        ver = v + ver
    if dashed:
        if not ver:
            return name, None       # no version
        else:
            return name, ver[1:]    # ver without leading '-'

    m = re.search(r'\.v.+$', filename)  # jpegsrc.v9d
    if m is not None:
        name = filename[:m.start()]
        ver  = filename[m.start()+1:]
        return name, ver

    m = re.match(r'(?P<name>.*[^0-9])(?P<rev>[0-9]+)$', filename) # IPAexfont00201
    if m is not None:
        return m.group('name'), m.group('rev')

    if not failonerr:
        return None

    raise RuntimeError('Unsupported url: %r' % (url,))


# isconf returns whether url points to data related to configuration file
# (contrary to binary executable), or a simple/small script.
def isconf(url):
    if url.endswith('.conf.in')                 or \
       url.endswith('.cfg')                     or \
       url.endswith('.cfg.in')                  or \
       url.endswith('.cnf.in')                  or \
       url.endswith('.cfg.jinja2')              or \
       url.endswith('.cfg.jinja2.in')           or \
       url.endswith('.jinja2.cfg')              or \
       url.endswith('.jinja2.library')          or \
       url.endswith('.asn')                     or \
       url.endswith('/ltelogs.jinja2.sh')       or \
       '/promise/' in url                       or \
       url.endswith('/templates/wrapper.in')    or \
       url.endswith('/logrotate_entry.in')      or \
       url.endswith('/matplotlibrc.in')         or \
       url.endswith('/goenv.sh.in')             :
        return True
    return False


# removeprefix/removesuffix provides fallbacks similarly-named str methods.
def removeprefix(s, prefix):
    if s.startswith(prefix):
        s = s[len(prefix):]
    return s
def removesuffix(s, suffix):
    if s.endswith(suffix):
        s = s[:len(s)-len(suffix)]
    return s


# spdx_license_canon returns canonical SPDX license ID for license_name.
#
# it is an error is license_name cannot be recognized.
_licdb = {  # spdx -> variants
    'Apache-2.0':   'Apache License 2.0 | Apache License, Version 2.0 | http://www.apache.org/licenses/LICENSE-2.0',
    'BSD-2-Clause': '2-Clause BSD',
    'BSD-3-Clause': '3-clause BSD License | BSD 3-Clause License | BSD (3-clause)',
    'CNRI-Python':  'Python License (CNRI Python License)',
    'EPL-1.0':      'Eclipse Public License 1.0',
    'HPND':         '',
    'ISC':          'ISC license',
    'LGPL-2.1':     '',
    'GPL-2.0+':     'GPL 2+',
    'GPL-3.0+':     'GNU GPLv3+',
    'GPL-3.0':      'GPLv3 | GPL3 | GNU General Public License Version 3',
    'MIT':          'MIT License',
    'MPL-2.0':      '',
    'ZPL-2.1':      'ZPL 2.1 (http://www.zope.org/Resources/License/ZPL-2.1)',

    'GPL-3.0+-with-NXD-exception':   'GPLv3+ with wide exception for FOSS | GPLv3+ with wide exception for Open-Source',

    # XXX no versions
    'BSD':          'BSD License | OSI Approved :: BSD License',
    'GPL':          '',
    'LGPL':         '',
    'PSF':          'PSF license | Python Software Foundation License',
}
_licrevdb = {}  # variant -> spdx
for lic, variants in _licdb.items():
    vv = []
    if variants:
        vv = variants.split('|')
    vv.append(lic)                      # MIT   -> MIT
    vv.append('"%s"' % lic)             # "MIT" -> MIT
    licsp = lic.replace('-', ' ')
    if licsp != lic:
        vv.append(licsp)                # ZPL 2.1 -> ZPL-2.1
    for v in vv:
        v = v.strip()
        assert v not in _licrevdb, repr(v)
        _licrevdb[v] = lic
def spdx_license_canon(license_name):
    # XXX split on ';' ',' 'and' 'or'
    # XXX split as if list ['MIT', 'BSD']  (pystemmer)
    lic = license_name.strip()
    lic = removeprefix(lic, "deprecated_")
    lic = removesuffix(lic, "-only")
    if lic.endswith('-or-later'):
        lic = removesuffix(lic, '-or-later')
        lic += '+'
    spdx = _licrevdb.get(lic)
    if spdx is None:
        return lic # XXX temp
        raise ValueError('unknown license %r' % license_name)
    return spdx


# license_of returns license of package at url.
def license_of(url):
    known = {
        # XXX hack - no suport for .lz
        'gettext-0.21.tar.lz':  'LGPL-2.1+ and GPL-3.0+',
        'expat-2.2.10.tar.lz':  'MIT',
        'wget-1.20.3.tar.lz':   'GPL-3.0+',

        'en-default.pyrnn.gz':  None,
        'http://geolite.maxmind.com/download/geoip/database/GeoLite2-Country.tar.gz':  None,

        'https://lab.nexedi.com/bk/onlyoffice_core/repository/archive.tar.bz2?ref=8a40eb47bd80a40ecde14c223525b21852d2fc9f':  'AGPL-3.0',

        '?p=config.git;a=snapshot;h=5e531d391852a54e7fab2d8ff55625fca514b305;sf=tgz': 'GPL-3.0-with-autoconf-exception',
        'https://sourceforge.net/projects/swig/files/swig/swig-3.0.12/swig-3.0.12.tar.gz/download':  'GPL-3.0+',

        'https://lab.nexedi.com/nexedi/erp5.git':           'GPL-3.0+-with-NXD-exception',
        'https://lab.nexedi.com/nexedi/erp5-bin.git':       'GPL-3.0+-with-NXD-exception',
        'https://lab.nexedi.com/nexedi/erp5-doc.git':       'GPL-3.0+-with-NXD-exception',
        'https://lab.nexedi.com/nexedi/neoppod.git':        'GPL-3.0+-with-NXD-exception',
        'https://lab.nexedi.com/nexedi/re6stnet.git':       'GPL-3.0+-with-NXD-exception',
        'https://lab.nexedi.com/nexedi/wendelin.core.git':  'GPL-3.0+-with-NXD-exception',
        'https://lab.nexedi.com/nexedi/wendelin.git':       'GPL-3.0+-with-NXD-exception',
        'https://lab.nexedi.com/nexedi/cloudooo.git':       'GPL-3.0+-with-NXD-exception',
    }
    if url in known:
        return known[url]
    _ = basename(url)
    if _ in known:
        return known[_]

    assert '//' in url, url
    _, xpath = url.split('//', 1)       # https://tukaani.org/xz/xz-5.2.5.tar.bz2 -> tukaani.org/xz/xz-5.2.5.tar.bz
    #assert '//' not in xpath, xpath    XXX archive.mariadb.org//mariadb-10.3.35/source/mariadb-10.3.35.tar.gz
    assert '..' not in xpath, xpath
    lpath = '.CACHE/lic/'+xpath
    xpath = '.CACHE/src/'+xpath
    if not exists(xpath):
        # download + unpack -> xpath
        pkgfile = '.CACHE/pkg/' + basename(xpath)
        if not exists(pkgfile):
            mkdir_p('.CACHE/pkg')
            wget(url, pkgfile)

        print('X %s ...' % pkgfile, file=sys.stderr)
        shutil.unpack_archive(pkgfile, xpath)

    _ = glob('%s/*' % xpath)        # many tarballs come with single directory inside
    if len(_) == 1:
        xpath = _[0]

    if not exists(lpath):
        print('A %s ...' % xpath, file=sys.stderr)
        licv = subprocess.check_output(['./go-license-detector/cmd/license-detector/license-detector', xpath])
        licv = licv.decode('utf-8')
        licv = licv.splitlines()
        assert licv[0] == xpath, (licv[0], xpath)
        licv = licv[1:]
        #print(licv, file=sys.stderr)
        lic = licv[0] # first best match
        lic = lic.strip()
        if lic == "no license file was found":
            lic = ''
        else:
            lic = lic.split(None, 1)[1]
        print('  -> %s' % lic, file=sys.stderr)
        mkdir_p(dirname(lpath))
        with open(lpath, 'w') as f:
            f.write(lic)
    else:
        with open(lpath, 'r') as f:
            lic = f.read()

    if lic == '':
        lic = None

    return lic


# wget downloads url to dstfile.
# it first tries shacache and the direct download.
_nc = None
def wget(url, dstfile):
    print('DL %s ...' % url,  file=sys.stderr)

    # see src/zc/buildout/download.py
    global _nc
    if _nc is None:
        _nc = NetworkcacheClient({
                    'download-cache-url': 'http://shacache.nxdcdn.com',
                    'download-dir-url':   'http://shadir.nxdcdn.com',
        })
    cachekey = 'file-urlmd5:' + md5(url.encode('utf-8')).hexdigest()
    if _nc.tryDownload(cachekey):
        with _nc:
            entry = next(_nc.select(cachekey, {'url': url}), None)
            if entry is not None:
                #print('DL %s ... (cache)' % url,  file=sys.stderr)
                with closing(_nc.download(entry['sha512'])) as src, \
                     open(dstfile+'.part', 'wb') as dst:
                    shutil.copyfileobj(src, dst)
                os.rename(dstfile+'.part', dstfile)
                return

    # direct
    print('DL %s ... (direct)' % url,  file=sys.stderr)
    1/0
    #urlretrieve(url, dstfile)


# XXX temp hack: disable ipv6
# https://stackoverflow.com/a/6319043/9456786
import socket
origGetAddrInfo = socket.getaddrinfo

def getAddrInfoWrapper(host, port, family=0, socktype=0, proto=0, flags=0):
    return origGetAddrInfo(host, port, socket.AF_INET, socktype, proto, flags)

# replace the original socket.getaddrinfo by our version
socket.getaddrinfo = getAddrInfoWrapper


def mkdir_p(path):
    os.makedirs(path, exist_ok=True)


# ----------------------------------------

# fmt_bom formats BOM into text.
def fmt_bom(bom): # -> str
    outv = []
    def emit(text):
        outv.append(text+'\n')

    # emit BOM grouped by kind
    kinds = set()
    for info in bom.values():
        kinds.add(info.kind)

    for kind in sorted(kinds):
        if kind != '':
            emit('\n>>> %ss:' % kind)
        for bkey in sorted(bom, key=lambda _: _[0].lower()):  # by name, ignoring case
            info = bom[bkey]
            if info.kind == kind:
                # TODO autoalign
                emit('%-28s %-10s %-20s %s' % (info.name, info.version, info.license, info.url))

    return ''.join(outv)

def main():
    if len(sys.argv) != 3  or  sys.argv[1] not in ('software', 'node'):
        print(__doc__, file=sys.stderr)
        sys.exit(2)

    logging.basicConfig(stream=sys.stderr, level=logging.INFO)

    what, arg = sys.argv[1:]
    if what == 'software':
        bom = bom_software(arg)
    elif what == 'node':
        bom = bom_node(arg)

    # print retrieved BOM
    # TODO also consider emitting machine-readable format, e.g. json if run with --json
    # TODO for json, consider schema from https://cyclonedx.org/specification/overview/
    print(fmt_bom(bom))


if __name__ == '__main__':
    main()
